package net.pharmatrak.babelj;

import java.io.*;
import java.net.*;
import java.util.*;
import java.lang.reflect.*;

/**
 * <blockquote><i>
 * Babelj is a language server library for Java.
 * Copyright (C) 1999  Pharmatrak
 * <p>
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * <p>
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 * <p>
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 * <p>
 * </i></blockquote>
 *
 * @author  Timothy W Macinta (twm@alum.mit.edu)
 **/

public class BabeljConnection implements Runnable {

  Socket sock = null;
  InputStream in;
  OutputStream out;

  boolean running = true;

  /**
   * This is where references are kept.
   * References need to be freed manually,
   * so the client is responsible for garbage
   * collection if the connection is going to
   * last awhile or there are going to be a large
   * number of references generated.  References
   * are automatically garbage collected
   * when the connection is closed, though.
   * I may add a way to have persistent references
   * later (you could do that too).
   **/
  Hashtable refs = new Hashtable();
  Hashtable reverse_refs = new Hashtable();
  long next_ref_id = 0;

  public BabeljConnection(Socket sock) throws IOException {
    this(sock.getInputStream(), sock.getOutputStream());
    this.sock = sock;
  }
  
  public BabeljConnection(InputStream in, OutputStream out) throws IOException {
    if (!(in instanceof BufferedInputStream)) in = new BufferedInputStream(in);
    if (!(out instanceof BufferedOutputStream)) out = new BufferedOutputStream(out);
    this.in = in;
    this.out = out;
    new Thread(this).start();
  }

  public void run() {
    boolean initialized = false;
    try {
      while (running) {
	String command_name = readCommandName();
	if (initialized) {
	  
	  // Can't re-set the version number

	  if (command_name.equalsIgnoreCase(Commands.BABELJ_VERSION)) {
	    returnError("Version confirmation has already occured.");
	    running = false;
	  } else if (command_name.equalsIgnoreCase(Commands.LOGOUT)) {

	    // Client is done with connection and logging out

	    returnString("Goodbye!");
	    running = false;
	  } else if (command_name.equalsIgnoreCase(Commands.METHOD_CALL)) {

	    // Client is calling an instance method

	    processMethodCall();
	  } else if (command_name.equalsIgnoreCase(Commands.STATIC_METHOD_CALL)) {

	    // Client is calling a static method

	    processStaticMethodCall();
	  } else if (command_name.equalsIgnoreCase(Commands.NEW_INSTANCE)) {

	    // Client is requesting a new instance

	    processNewInstanceCall();
	  } else if (command_name.equalsIgnoreCase(Commands.GET_FIELD)) {

	    // Client is requesting a reference to an instance field

	    processGetFieldCall();
	  } else if (command_name.equalsIgnoreCase(Commands.GET_STATIC_FIELD)) {

	    // Client is requesting a new instance

	    processGetStaticFieldCall();
	  } else if (command_name.equalsIgnoreCase(Commands.FREE_REF)) {

	    // Client is requesting that a reference be freed

	    processFreeRefCall();
	  }

	} else {
	  
	  // We haven't been initialized yet, so look
	  // for a version number

	  if (command_name.equalsIgnoreCase(Commands.BABELJ_VERSION)) {
	    String ver = readString();
	    if (!ver.equalsIgnoreCase(Constants.VERSION_0_1)) {
	      returnError("Version number not supported.  The only version supported by this server is \""+Constants.VERSION_0_1+"\" - you may need to upgrade (or potentially downgrade) your server.");
	      running = false;
	    }
	    returnString("Welcome!");
	    initialized = true;
	  } else {
	    returnError("Version number must be sent first.");
	    running = false;
	  }
	}
	out.flush();
      }
    } catch (IOException e) {
      e.printStackTrace();
    }
    cleanup();
  }

  void cleanup() {
    try {
      in.close();
      out.close();
    } catch (IOException e) {}
    if (sock != null) {
      try {
	sock.close();
      } catch (IOException e2) {}
    }
  }    

  void processMethodCall() throws IOException {
    try {
      Object obj = readObject();
      String method_name = (String) readObject();
      Object args[] = (Object[]) readObject();
      try {
	Method method = findMethod(method_name, args, obj.getClass().getMethods());
	if (method == null) {
	  returnError("No such method.");
	  return;
	}
	returnObject(method.invoke(obj, args));
	return;
      } catch (Exception e) {
	returnException(e);
	return;
      }
    } catch (IllegalArgumentException e2) {
      returnException(e2);
      stop();
      return;
    }
  }

  /** Picks out the method that matches the argument types. **/
  Method findMethod(String method_name, Object args[], Method m[]) {
    for (int i = 0; i < m.length; i++) {
      if (m[i].getName().equals(method_name)) {
	Class[] pt = m[i].getParameterTypes();
	if (pt.length == args.length) {
	  boolean matched = true;
	  for (int x = 0; x < pt.length; x++) {
	    if (!pt[x].isInstance(args[x])) {
	      try {
		if (pt[x].getName().indexOf('.') < 0) {
		  Field f = args[x].getClass().getField("TYPE");
		  if (f.get(null).equals(pt[x])) continue;
		}
	      } catch (NoSuchFieldException e) {
	      } catch (IllegalAccessException e2) {}
	      matched = false;
	      break;
	    }
	  }
	  if (matched) {
	    return m[i];
	  }
	}
      }
    }

    // this method should do some smart things like
    // attempting to parse Strings into numbers where
    // appropriate before returning null and replacing
    // the corresponding array elements

    return null;
  }

  /** Picks out the constructor that matches the argument types. **/
  Constructor findConstructor(Object args[], Constructor con[]) {
    for (int i = 0; i < con.length; i++) {
      Class[] pt = con[i].getParameterTypes();
      if (pt.length == args.length) {
	boolean matched = true;
	for (int x = 0; x < pt.length; x++) {
	  if (!pt[x].isInstance(args[x])) {
	    try {
	      if (pt[x].getName().indexOf('.') < 0) {
		Field f = args[x].getClass().getField("TYPE");
		if (f.get(null).equals(pt[x])) continue;
	      }
	    } catch (NoSuchFieldException e) {
	    } catch (IllegalAccessException e2) {}
	    matched = false;
	    break;
	  }
	}
	if (matched) {
	  return con[i];
	}
      }
    }

    // this method should do some smart things like
    // attempting to parse Strings into numbers where
    // appropriate before returning null and replacing
    // the corresponding array elements

    return null;
  }

  void processStaticMethodCall() throws IOException {
    try {
      String classname = (String) readObject();
      String method_name = (String) readObject();
      Object args[] = (Object[]) readObject();
      try {
	Method method = findMethod(method_name, args, Class.forName(classname).getMethods());
	if (method == null) {
	  returnError("No such method.");
	  return;
	}
	returnObject(method.invoke(null, args));
	return;
      } catch (Exception e) {
	returnException(e);
	return;
      }
    } catch (IllegalArgumentException e2) {
      returnException(e2);
      stop();
      return;
    }
  }

  void processNewInstanceCall() throws IOException {
    try {
      String classname = (String) readObject();
      Object args[] = (Object[]) readObject();
      try {
	Constructor con = findConstructor(args, Class.forName(classname).getConstructors());
	if (con == null) {
	  returnError("No such constructor.");
	  return;
	}
	returnObject(con.newInstance(args));
	return;
      } catch (Exception e) {
	returnException(e);
	return;
      }
    } catch (IllegalArgumentException e2) {
      returnException(e2);
      stop();
      return;
    }
  }

  void processGetFieldCall() throws IOException {
    try {
      Object obj = readObject();
      String field_name = (String) readObject();
      try {
	Field f = obj.getClass().getField(field_name);
	returnObject(f.get(obj));
	return;
      } catch (Exception e) {
	returnException(e);
	return;
      }
    } catch (IllegalArgumentException e2) {
      returnException(e2);
      stop();
      return;
    }
  }

  void processGetStaticFieldCall() throws IOException {
    try {
      String classname = (String) readObject();
      String field_name = (String) readObject();
      try {
	Field f = Class.forName(classname).getField(field_name);
	returnObject(f.get(null));
	return;
      } catch (Exception e) {
	returnException(e);
	return;
      }
    } catch (IllegalArgumentException e2) {
      returnException(e2);
      stop();
      return;
    }
  }

  Object readObject() throws IOException {
    String object_type = readObjectType();
    if (object_type.equalsIgnoreCase(ObjectTypes.NULL)) {
      return null;
    } else if (object_type.equalsIgnoreCase(ObjectTypes.VOID)) {
      throw new IllegalArgumentException("Void is not a valid method argument!");
    } else if (object_type.equalsIgnoreCase(ObjectTypes.EXCEPTION)) {
      throw new IllegalArgumentException("Method callers can't throw exceptions!");
    } else if (object_type.equalsIgnoreCase(ObjectTypes.VAR)) {
      int len = Integer.parseInt(readString());
      return attemptStringConversion(readString(len));
    } else if (object_type.equalsIgnoreCase(ObjectTypes.STRING)) {
      int len = Integer.parseInt(readString());
      return readString(len);
    } else if (object_type.equalsIgnoreCase(ObjectTypes.INTEGER)) {
      return new Integer(readString());
    } else if (object_type.equalsIgnoreCase(ObjectTypes.LONG)) {
      return new Long(readString());
    } else if (object_type.equalsIgnoreCase(ObjectTypes.DOUBLE)) {
      return new Double(readString());
    } else if (object_type.equalsIgnoreCase(ObjectTypes.FLOAT)) {
      return new Float(readString());
    } else if (object_type.equalsIgnoreCase(ObjectTypes.BYTE)) {
      return new Byte(readString());
    } else if (object_type.equalsIgnoreCase(ObjectTypes.SHORT)) {
      return new Short(readString());
    } else if (object_type.equalsIgnoreCase(ObjectTypes.BOOLEAN)) {
      return new Boolean(readString());
    } else if (object_type.equalsIgnoreCase(ObjectTypes.ARRAY)) {
      int len = Integer.parseInt(readString());
      Object ar[] = new Object[len];
      for (int i = 0; i < ar.length; i++) {
	ar[i] = readObject();
      }
      return ar;
    } else if (object_type.equalsIgnoreCase(ObjectTypes.REF)) {
      String s = readString();
      synchronized(refs) {
	return refs.get(s);
      }
    } else {
      throw new IllegalArgumentException("Unrecognized type \""+object_type+"\"");
    }
  }

  /**
   * Attempts to convert a String to a wrapped primitive type.
   * I have attempted to convert Strings to primitives only when the
   * conversion back to a string would yeild the same String.
   **/
  Object attemptStringConversion(String s) {
    if (s.equals("true")) return Boolean.TRUE;
    if (s.equals("false")) return Boolean.TRUE;

    // OK, throwing an exception as the common case
    // isn't terribly efficeint... optimize if the need arises

    try {
      Double d = new Double(s);
      if (d.longValue() == d.doubleValue())
	return new Long(d.longValue());
      return d;
    } catch (NumberFormatException e) {
    }
    return s;
  }

  /**
   * Reads a String from in.  Reading is terminated when
   * a '\n' is encountered.
   **/
  String readString() throws IOException {
    //System.out.println("BabeljConnection.readString() 1");
    int c;
    StringBuffer sb = new StringBuffer();
    while ((c = in.read()) != '\n') {
      if (c < 0) throw new EOFException();
      sb.append((char) c);
    }
    //System.out.println("BabeljConnection.readString() 2");
    return new String(sb);
  }

  /**
   * Reads exactly strlen bytes into a String.
   **/
  String readString(int strlen) throws IOException {
    //System.out.println("BabeljConnection.readString("+strlen+") 1");
    int c = 0;
    StringBuffer sb = new StringBuffer();
    while (strlen-- > 0 && (c = in.read()) >= 0) {
      sb.append((char) c);
    }
    if (c < 0) throw new EOFException();
    //System.out.println("BabeljConnection.readString("+strlen+") 2");
    return new String(sb);
  }

  String readCommandName() throws IOException {
    return readString();
  }

  public void stop() {
    running = false;
  }

  void returnException(Exception e) throws IOException {
    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    PrintWriter tout = new PrintWriter(new OutputStreamWriter(baos));
    e.printStackTrace(tout);
    tout.flush();
    tout.close();
    writeString(ObjectTypes.EXCEPTION+'\n');
    returnString(new String(baos.toByteArray()));
  }

  void returnError(String error_message) throws IOException {
    returnException(new Exception(error_message));
  }

  void returnString(String str) throws IOException {
    writeString(ObjectTypes.STRING+'\n');
    writeString(str.length()+"\n");
    writeString(str);
  }
  
  void writeString(String str) throws IOException {
    out.write(str.getBytes());
  }

  void returnObject(Object obj) throws IOException {
    if (obj == null) {
      writeString(ObjectTypes.NULL+'\n');
      return;
    }
    if (obj instanceof Void) {
      writeString(ObjectTypes.VOID+'\n');
      return;
    }
    if (obj instanceof String) {
      returnString((String) obj);
      return;
    }
    if (obj instanceof Integer) {
      writeString(ObjectTypes.INTEGER+'\n'+obj+'\n');
      return;
    }
    if (obj instanceof Long) {
      writeString(ObjectTypes.LONG+'\n'+obj+'\n');
      return;
    }
    if (obj instanceof Double) {
      writeString(ObjectTypes.DOUBLE+'\n'+obj+'\n');
      return;
    }
    if (obj instanceof Float) {
      writeString(ObjectTypes.FLOAT+'\n'+obj+'\n');
      return;
    }
    if (obj instanceof Byte) {
      writeString(ObjectTypes.BYTE+'\n'+obj+'\n');
      return;
    }
    if (obj instanceof Short) {
      writeString(ObjectTypes.SHORT+'\n'+obj+'\n');
      return;
    }
    if (obj instanceof Boolean) {
      writeString(ObjectTypes.BOOLEAN+'\n'+obj+'\n');
      return;
    }

    // Not a supported type so pass by reference

    returnRef(obj);
    return;
  }

  void returnBoolean(boolean val) throws IOException {
    writeString(ObjectTypes.BOOLEAN+'\n'+val+'\n');
  }

  void returnRef(Object obj) throws IOException {
    String rid;
    synchronized(refs) {
      rid = (String) reverse_refs.get(obj);
      if (rid == null) {
	rid = "r"+next_ref_id++;
	refs.put(rid, obj);
	reverse_refs.put(obj, rid);
      }
    }
    writeString(ObjectTypes.REF + '\n' + rid + '\n');
  }

  String readObjectType() throws IOException {
    return readString();
  }
  
  void processFreeRefCall() throws IOException {
    try {
      Object obj = readObject();
      String rid;
      synchronized(refs) {
	rid = (String) reverse_refs.remove(obj);
	if (rid != null) {
	  refs.remove(rid);
	}
      } 
      returnBoolean(rid != null);
    } catch (IllegalArgumentException e2) {
      returnException(e2);
      stop();
      return;
    }
  }

}
